/*
   emu5110
	-------
	An emulator for the IBM 5110
	(C) 2007-2008 Christian Corti
	
	Emulator core
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "emu.h"
#include "disk_io.h"
#include "printer_io.h"
#include "tape_io.h"

#define DEBUG 0

/* select default mode */
/*#define APL*/
#define BASIC

#if defined(APL) && defined(BASIC)
#error Only one of APL or BASIC may be defined!
#endif

/* file names of the individual ROS images */
#define EXECROS_NAME		"execros.bin"
#define COMMONROS_NAME	"comros.bin"
#define BASNXROS_NAME	"basicnx.bin"
#define BASROS_NAME		"basicros.bin"
#define APLNXROS_NAME	"aplnx.bin"
#define APLROS_NAME		"aplros.bin"

void emu_reset(void);
void emu_start(void);
int emu_fetch(void);
int emu_do(void);
void emu_keyboard(UBYTE code);
void emu_dump_mem(void);
int emu_main(void);

void trace_apl(UBYTE b, int x);
void trace_basic(UBYTE b, int x);
void dump_regs(void);
int load_ros(void);

void check_int(void);
void Opcode0(void);
void OpcodeC(void);
void OpcodeE(void);
void DoCTRL(int dev, int imm);
void DoPUTB(int dev, UBYTE c);
UBYTE DoGETB(int dev);

#define INT_ENA	0x80
#define INT_3		0x04
#define INT_2		0x02
#define INT_1		0x01

enum {
 ERR_NOERR=0,
 ERR_RWS_ALLOC,
 ERR_EXECROS_FILE, ERR_EXECROS_ALLOC, ERR_EXECROS_READ,
 ERR_COMMONROS_FILE, ERR_COMMONROS_ALLOC, ERR_COMMONROS_READ,
 ERR_APLROS_FILE, ERR_APLROS_ALLOC, ERR_APLROS_READ,
 ERR_APLNXROS_FILE, ERR_APLNXROS_ALLOC, ERR_APLNXROS_READ,
 ERR_BASROS_FILE, ERR_BASROS_ALLOC, ERR_BASROS_READ,
 ERR_BASNXROS_FILE, ERR_BASNXROS_ALLOC, ERR_BASNXROS_READ,
};

enum {
 MODE_RWS=0, MODE_BUP, MODE_ROS
};

#define POSTCREMENT \
	if(n3<4) \
		curr_regs[n2] += n3+1; \
	else if(n3<8) \
		curr_regs[n2] -= n3-3;

#ifdef LITTLE_ENDIAN
#define REGHI(reg)	curr_regsb[(reg<<1)+1]
#define REGLO(reg)	curr_regsb[(reg<<1)]
#else
#define REGLO(reg)	curr_regsb[(reg<<1)+1]
#define REGHI(reg)	curr_regsb[(reg<<1)]
#endif

#define IMM				(OP&0xFF)
#define INC_PC			curr_regs[0]+=2

#if DEBUG
#ifdef APL
#define TRACE(a,b)	trace_apl(a,b)
#else
#define TRACE(a,b)	trace_basic(a,b)
#endif
#else
#define TRACE(a,b)
#endif

/* The individual Executable ROS areas and the RWS */
USHORT *RWS, *ExecROS, *APLROS, *BASROS;
USHORT *curr_ros, *language_ros;
UBYTE *RWSb;					/* RWS with byte addressing */

/* The non-executable ROS areas */
USHORT *CommonROS, *BASNXROS, *APLNXROS;
USHORT *nxros, nxros_addr, nxros_waddr;
int nxros_rtoggle, nxros_wtoggle, nxros_xtoggle;

int emuvar_display;
unsigned long emuvar_timer;

/* Interrupt handling */
int int_mask, int1, int2, int3;
UBYTE kbdcode;

/* Registers */
USHORT *curr_regs;			/* current level register set; word... */
UBYTE *curr_regsb;			/* ... and byte addressing */

/* Instruction handling */
USHORT OP;						/* current opcode */
UBYTE n1, n2, n3;				/* last three nibbles of opcode */
int level, halt, mode[4];

/* CPU dump */
int dump;
FILE *dumpfile;

/********************/

void emu_reset()
{
	level = 0;
	curr_ros = ExecROS;
	mode[0] = MODE_BUP;	/* in 'bring up program' mode */
	mode[1] = MODE_ROS;
	mode[2] = MODE_ROS;
	mode[3] = MODE_ROS;

	halt = FALSE;

	nxros = CommonROS;
	nxros_rtoggle = 0;
	nxros_wtoggle = 0;
	nxros_xtoggle = 0;

	int_mask = INT_ENA|INT_1|INT_2|INT_3;
	int1 = FALSE;
	int2 = FALSE;
	int3 = FALSE;
	
	curr_regs  = RWS;
	curr_regsb = (UBYTE *)RWS;
#ifdef LITTLE_ENDIAN
	curr_regs[0] = (((UBYTE *)ExecROS)[0])<<8 | ((UBYTE *)ExecROS)[1];
#else
	curr_regs[0] = ExecROS[0];
#endif

	emuvar_display = 1;		/* display on */
	emuvar_timer = 0;
}

int emu_fetch()
{
#ifdef LITTLE_ENDIAN
	USHORT swptmp;
#endif

	check_int();		/* check for pending interrupts */

	OP = ((mode[level]==MODE_RWS)?RWS:(level?ExecROS:curr_ros))[curr_regs[0]>>1];
#ifdef LITTLE_ENDIAN
	OP = SWAB(OP);
#endif
	n1 = (OP&0x0F00)>>8;
	n2 = (OP&0x00F0)>>4;
	n3 = (OP&0x000F);

	/* some kludges */
	if(!level && mode[0]==MODE_ROS && curr_ros==ExecROS)
	{
		if(curr_regs[0]==0x42CC)
		{
			/* ring the bell when the Screen I/O routine "beep" is called */
			fprintf(stderr, "\007");
		}
		else if(curr_regs[0]==0x4A00)
		{
			/* the Printer I/O supervisor is comletely emulated */
			PrinterIO(curr_regs[3]);
			curr_regs[0] = 0x5142;
		}
	}

	INC_PC;
	emuvar_timer++;

	switch(OP>>12)
	{
		case 0:
			Opcode0();
			break;

		case 1:		/* CTRL */
			DoCTRL(n1, IMM);
			break;
			
		case 2:		/* MOVE Rx, i' */
#ifdef LITTLE_ENDIAN
			if(IMM < 0x0040)
				curr_regs[n1] = RWS[IMM];
			else
				curr_regs[n1] = SWAB(RWS[IMM]);
#else
			curr_regs[n1] = RWS[IMM];
#endif
			break;
			
		case 3:		/* MOVE i', Rx */
#ifdef LITTLE_ENDIAN
			if(IMM < 0x0040)
				RWS[IMM] = curr_regs[n1];
			else
				RWS[IMM] = SWAB(curr_regs[n1]);
#else
			RWS[IMM] = curr_regs[n1];
#endif

			break;
			
		case 4:		/* PUTB i, (Rx)* */
#ifdef LITTLE_ENDIAN
			swptmp = curr_regs[n2];
			if(swptmp < 0x0080)
				swptmp ^= 1;
			DoPUTB(n1, RWSb[swptmp]);
#else
			DoPUTB(n1, RWSb[curr_regs[n2]]);
#endif
			POSTCREMENT
			break;
			
		case 5:		/* MOVE (Ry)*, Rx */
#ifdef LITTLE_ENDIAN
			swptmp = curr_regs[n2];
			if(swptmp < 0x0080)
				RWS[swptmp>>1] = curr_regs[n1];
			else
				RWS[swptmp>>1] = SWAB(curr_regs[n1]);
#else
			RWS[curr_regs[n2]>>1] = curr_regs[n1];
#endif
			POSTCREMENT
			break;
			
		case 6:		/* MOVB Rx, (Ry)* */
#ifdef LITTLE_ENDIAN
			swptmp = curr_regs[n2];
			if(swptmp < 0x0080)
				swptmp ^= 1;
			curr_regs[n1] = (USHORT)(RWSb[swptmp]);
#else
			/* This is correct, since Hi(Rx) must be set to 0 */
			curr_regs[n1] = (USHORT)(RWSb[curr_regs[n2]]);
#endif
			if(n1 != n2)
				POSTCREMENT
			break;

		case 7:		/* MOVB (Ry)*, Rx */
#ifdef LITTLE_ENDIAN
			swptmp = curr_regs[n2];
			if(swptmp < 0x0080)
				swptmp ^= 1;
			RWSb[swptmp] = REGLO(n1);
#else
			RWSb[curr_regs[n2]] = REGLO(n1);
#endif
			POSTCREMENT
			break;
			
		case 8:		/* LBI Rx, i */
			REGLO(n1) = IMM;
			break;
			
		case 9:		/* CLR Rx, i */
			REGLO(n1) &= ~IMM;
			break;
			
		case 10:		/* ADD Rx, i */
			curr_regs[n1] += IMM+1;
			break;
			
		case 11:		/* SET Rx, i */
			REGLO(n1) |= IMM;
			break;
			
		case 12:		/* Skip instructions */
			OpcodeC();
			break;
			
		case 13:		/* MOVE Rx, (Ry)* */
#ifdef LITTLE_ENDIAN
			swptmp = curr_regs[n2];
			if(swptmp < 0x0080)
				curr_regs[n1] = RWS[swptmp>>1];
			else
				curr_regs[n1] = SWAB(RWS[swptmp>>1]);
#else
			curr_regs[n1] = RWS[curr_regs[n2]>>1];
#endif
			if(n1 != n2)
				POSTCREMENT
			break;
			
		case 14:		/* STAT or Shift/Rotate */
			OpcodeE();
			break;
			
		case 15:		/* SUB Rx, i */
			curr_regs[n1] -= (IMM+1);
			break;
	}

	TRACE(0, 0);

	return halt;
}

void trace_apl(UBYTE b, int x)
{
	if(!dump || level || mode[0]!=MODE_ROS || curr_ros!=APLROS)
		return;

	if(curr_regs[0]==0x01FE)
	{
		fprintf(dumpfile, "\n%05X(%04X):", nxros_addr<<1, nxros_addr);
	}
	else if(curr_regs[0]==0x020E)
	{
		fprintf(dumpfile, "\n%05X(RWS): %02X %02X", curr_regs[14]-2,
				REGHI(4), REGLO(4));
	}
	else if(curr_regs[0]==0x0208 || curr_regs[0]==0x0218)
	{
		fprintf(dumpfile, " (->%04X)", SWAB(RWS[curr_regs[2]>>1]));
	}
	else if(curr_regs[0] == 0x03D4)
	{
		fprintf(dumpfile, "\n***Access Exception:");
		dump_regs();
	}
	else if(curr_regs[0] == 0x03C0)
	{
		fprintf(dumpfile, "\n***Data Exception:");
		dump_regs();
	}
	else if(x)
	{
		fprintf(dumpfile, " %02X", b);
	}
}

void trace_basic(UBYTE b, int x)
{
	if(!dump || level || mode[0]!=MODE_ROS || curr_ros!=BASROS)
		return;

	if(curr_regs[0] == 0x019A)
	{
		fprintf(dumpfile, "\n%05X(%04X):", nxros_addr<<1, nxros_addr);
	}
	else if(x)
	{
		fprintf(dumpfile, " %02X", b);
		dump_regs();
	}
}

void check_int()
{
	DiskIRQ();
	TapeIRQ();
/*	PrinterIRQ(); */

	if(int_mask&INT_ENA && (int1||int2||int3))
	{
		/* interrupt occured, select level by priority */

		if(int3 && int_mask&INT_3)
		{
			level = 3;
		}
		else if(int2 && int_mask&INT_2)
		{
			level = 2;
		}
		else if(int1 && int_mask&INT_1)
		{
			level = 1;
		}

		curr_regs = &RWS[level<<4];
		curr_regsb = (UBYTE *)curr_regs;
	}
	else if(level != 0)
	{
		/* interrupt cleared, switch back to level 0 */
		level = 0;
		curr_regs = RWS;
		curr_regsb = (UBYTE *)RWS;
	}
}

inline void emu_start(){ halt=FALSE; }

int emu_do()
{
	int i;
	
	i = 0;
	while(i++<50 && !emu_fetch());

	return halt;
}		

void dump_regs()
{
	int i;
	FILE *x;
	
	if(dump)
		x=dumpfile;
	else
		x=stderr;
		
	fprintf(x, "\n");
#ifdef LITTLE_ENDIAN
	for(i=0; i<8; i++)
		fprintf(x, "R%02d:%02X%02X  ", i, REGHI(i), REGLO(i));
	fprintf(x, "\n");
	for(i=8; i<16; i++)
		fprintf(x, "R%02d:%02X%02X  ", i, REGHI(i), REGLO(i));
	fprintf(x, "\n");
#else
	for(i=0; i<8; i++)
		fprintf(x, "R%02d:%04X  ", i, curr_regs[i]);
	fprintf(x, "\n");
	for(i=8; i<16; i++)
		fprintf(x, "R%02d:%04X  ", i, curr_regs[i]);
	fprintf(x, "\n");
#endif
}

void Opcode0()
{
	UBYTE tmp;
	
	switch(n3)
	{
		case 0:		/* DEC2 */
			curr_regs[n1] = curr_regs[n2] - 2;
			break;
		case 1:		/* DEC */
			curr_regs[n1] = curr_regs[n2] - 1;
			break;
		case 2:		/* INC */
			curr_regs[n1] = curr_regs[n2] + 1;
			break;
		case 3:		/* INC2 */
			curr_regs[n1] = curr_regs[n2] + 2;
			break;
		case 4:		/* MOVE Rx, Ry */
			curr_regs[n1] = curr_regs[n2];
			break;
		case 5:		/* AND Rx, Ry */
			REGLO(n1) &= REGLO(n2);
			break;
		case 6:		/* OR Rx, Ry */
			REGLO(n1) |= REGLO(n2);
			break;
		case 7:		/* XOR Rx, Ry */
			REGLO(n1) ^= REGLO(n2);
			break;
		case 8:		/* ADD Rx, Ry */
			curr_regs[n1] += (UBYTE)REGLO(n2);
			break;
		case 9:		/* SUB Rx, Ry */
			curr_regs[n1] -= (USHORT)REGLO(n2);
			break;
		case 10:		/* ADDH Rx, Ry */
			curr_regs[n1] = (USHORT)REGHI(n1) + (USHORT)REGLO(n2);
			break;
		case 11:		/* ADDH2 Rx, Ry */
			curr_regs[n1] = (USHORT)REGHI(n1) + (USHORT)REGLO(n2);
			curr_regs[n1] -= 0x100;
			break;
		case 12:		/* MHL Rx, Ry */
			REGLO(n1) = REGHI(n2);
			break;
		case 13:		/* MLH Rx, Ry */
			REGHI(n1) = REGLO(n2);
			break;
		case 14:		/* GETB Ry, i */
			REGLO(n2) = DoGETB(n1);
			break;
		case 15:		/* GETADD Ry, i */
			tmp = DoGETB(n1);
			if((tmp&0xFE) == 0xFE)
			{
				/* add zero */
			}
			else if((tmp&0xFE) == 0xFC)
				curr_regs[n2] += 2;
			else if((tmp&0xFC) == 0xF8)
				curr_regs[n2] += 4;
			else if((tmp&0xF8) == 0xF0)
				curr_regs[n2] += 6;
			else if((tmp&0xF0) == 0xE0)
				curr_regs[n2] += 8;
			else if((tmp&0xE0) == 0xC0)
				curr_regs[n2] += 10;
			else if((tmp&0xC0) == 0x80)
				curr_regs[n2] += 12;
			else if((tmp&0x80) == 0x00)
				curr_regs[n2] += 14;

			break;
	}
}

void OpcodeC()
{
	switch(n3)
	{
		case 0:		/* SLE */
			if(REGLO(n1) <= REGLO(n2))
				INC_PC;
			else
				curr_regs[1] = curr_regs[0] + 2;
			break;
			
		case 1:		/* SLT */
			if(REGLO(n1) < REGLO(n2))
				INC_PC;
			else
				curr_regs[1] = curr_regs[0] + 2;
			break;
			
		case 2:		/* SE */
			if(REGLO(n1) == REGLO(n2))
				INC_PC;
			else
				curr_regs[1] = curr_regs[0] + 2;
			break;
			
		case 3:		/* SZ */
			if(REGLO(n1) == 0)
				INC_PC;
			else
				curr_regs[1] = curr_regs[0] + 2;
			break;
			
		case 4:		/* SS */
			if(REGLO(n1) == 0xFF)
				INC_PC;
			else
				curr_regs[1] = curr_regs[0] + 2;
			break;
			
		case 5:		/* SBS */
			if((REGLO(n1)&REGLO(n2)) == REGLO(n2))
				INC_PC;
			else
				curr_regs[1] = curr_regs[0] + 2;
			break;

		case 6:		/* SBC */
			if((REGLO(n1)&REGLO(n2)) == 0)
				INC_PC;
			else
				curr_regs[1] = curr_regs[0] + 2;
			break;

		case 7:		/* SBSH */
			if((REGHI(n1)&REGLO(n2)) == REGLO(n2))
				INC_PC;
			else
				curr_regs[1] = curr_regs[0] + 2;
			break;

		case 8:		/* SGT */
			if(REGLO(n1) > REGLO(n2))
				INC_PC;
			else
				curr_regs[1] = curr_regs[0] + 2;
			break;
			
		case 9:		/* SGE */
			if(REGLO(n1) >= REGLO(n2))
				INC_PC;
			else
				curr_regs[1] = curr_regs[0] + 2;
			break;
			
		case 10:		/* SNE */
			if(REGLO(n1) != REGLO(n2))
				INC_PC;
			else
				curr_regs[1] = curr_regs[0] + 2;
			break;
			
		case 11:		/* SNZ */
			if(REGLO(n1) != 0)
				INC_PC;
			else
				curr_regs[1] = curr_regs[0] + 2;
			break;
			
		case 12:		/* SNS */
			if(REGLO(n1) != 0xFF)
				INC_PC;
			else
				curr_regs[1] = curr_regs[0] + 2;
			break;
			
		case 13:		/* SNBS */
			if((REGLO(n1)&REGLO(n2)) != REGLO(n2))
				INC_PC;
			else
				curr_regs[1] = curr_regs[0] + 2;
			break;

		case 14:		/* SNBC */
			if((REGLO(n1)&REGLO(n2)) != 0)
				INC_PC;
			else
				curr_regs[1] = curr_regs[0] + 2;
			break;

		case 15:		/* SNBSH */
			if((REGHI(n1)&REGLO(n2)) != REGLO(n2))
				INC_PC;
			else
				curr_regs[1] = curr_regs[0] + 2;
			break;
	}
}

void OpcodeE()
{
	if(n1 == 0)
	{
		/* Shift/Rotate */
		switch(n3)
		{
			case 12:		/* SHR Rx */
				REGLO(n2) = (REGLO(n2)>>1) | (REGHI(n2)<<7);
				break;
				
			case 13:		/* ROR Rx */
				REGLO(n2) = (REGLO(n2)>>1) | (REGLO(n2)<<7);
				break;
				
			case 14:		/* ROR3 Rx */
				REGLO(n2) = (REGLO(n2)>>3) | (REGLO(n2)<<5);
				break;
			
			case 15:		/* SWAP Rx */
				REGLO(n2) = (REGLO(n2)>>4) | (REGLO(n2)<<4);
		}
	}
	else
	{
		/* GETB/STAT */

		if(n3 >= 0xC)
		{
			/* STAT */
			switch(n1)
			{
				case 1:	/* Common/Language ROS */
					if(!nxros_xtoggle)
						REGLO(n2) = nxros_addr >> 8;
					else
						REGLO(n2) = nxros_addr & 0xFF;
					nxros_xtoggle = 1-nxros_xtoggle;
					break;
					
				case 2:	/* Executable ROS */
					if(REGLO(n2) == 0x80)
						REGLO(n2) = 0xF0;
					else if(REGLO(n2) == 0x40)
						REGLO(n2) = 0xC0;
					else if(REGLO(n2) == 0x20)
						REGLO(n2) = 0x80;
					break;
					
				case 4:	/* Keyboard */
					if(REGLO(n2) == 0x80)
					{
						/* Read status */
						if(language_ros == BASROS)
							REGLO(n2) = 0x76;
						else
							REGLO(n2) = 0x36;

						if(kbdcode)
							REGLO(n2) |= 0x08;
					}
					else if(REGLO(n2) == 0x40)
					{
						/* Read scancode */
						REGLO(n2) = kbdcode;
						kbdcode = 0;
					}
					break;
					
				case 5:	/* Printer */
					REGLO(n2) = PrinterSTAT();
					break;
					
				case 13:	/* Diskette */
					REGLO(n2) = DiskSTAT();
					break;
					
				case 14:	/* Tape */
					REGLO(n2) = TapeSTAT(REGLO(n2));
					break;
					
				default:
					REGLO(n2) = 0xFF;
			}
		}
		else
		{
			/* GETB */
#ifdef LITTLE_ENDIAN
			USHORT swptmp;

			swptmp = curr_regs[n2];
			if(swptmp < 0x0080)
				swptmp ^= 1;
			RWSb[swptmp] = DoGETB(n1);
#else
			RWSb[curr_regs[n2]] = DoGETB(n1);
#endif
			POSTCREMENT
		}
	}
}

void DoCTRL(int dev, int imm)
{
	switch(dev)
	{
		case 0:	/* Processor/Display */
			if(!(imm&0x40))
				int_mask &= ~INT_ENA;
			else if(!(imm&0x20))
				int_mask |= INT_ENA;

			if(!(imm&0x10))
				emuvar_display = 0;
			else if(!(imm&0x08))
				emuvar_display = 1;
				
			if(!(imm&4))
			{
				if(mode[0] == MODE_ROS)
				{
					mode[0] = MODE_RWS;
				}
				else
				{
					/* We're either in MODE_BUP or MODE_RWS */
					mode[0] = MODE_ROS;
				}
			}
			break;
			
		case 1:	/* Common/Language ROS */
			if((imm & 2)==2)
				nxros = CommonROS;
			else if((imm & 4)==4)
				nxros = APLNXROS;
			else if((imm & 8)==8)
				nxros = BASNXROS;
			nxros_rtoggle = 0;
			nxros_wtoggle = 0;
			nxros_xtoggle = 0;
			break;
			
		case 2:	/* Executable ROS */
			if(imm == 0xBF)
			{
				/* Switch to interpreter ROS */
				curr_ros = language_ros;
			}
			else if(imm == 0x7F)
			{
				/* Switch to Executable ROS */
				curr_ros = ExecROS;
			}
			break;
			
		case 4:	/* Keyboard */
			if(!(imm&0x40))
				int3 = FALSE;
			if(!(imm&0x01))
				int_mask |= INT_3;
			else
				int_mask &= ~INT_3;
			break;
			
		case 5:	/* Printer */
			PrinterCTRL(imm);
			break;
			
		case 13:	/* Diskette */
			DiskCTRL(imm);
			break;
			
		case 14:	/* Tape */
			TapeCTRL(imm);
			break;
			
		case 15:	/* Adapter reset */
			if(imm & 0x80)
			{
				/* Async. communications/Serial I/O */
			}
			if(imm & 0x40)
			{
				/* Tape adapter */
				TapeReset();
			}
			if(imm & 0x20)
			{
				/* Keyboard */
				int3 = FALSE;
				int_mask &= ~INT_3;
				kbdcode = 0;
			}
			if(imm & 0x10)
			{
				/* Printer */
				PrinterReset();
			}
			if(imm & 0x08)
			{
				/* Display adapter */
				emuvar_display = 1;
			}
			if(imm & 0x04)
			{
				/* Device B */
			}
			if(imm & 0x02)
			{
				/* Device C */
			}
			if(imm & 0x01)
			{
				/* Diskette adapter */
				DiskReset();
			}
			break;
	}
}

void DoPUTB(int dev, UBYTE c)
{
	switch(dev)
	{
		case 0:	/* Processor */
			if(c&0x40)
				mode[1] = MODE_RWS;
			else
				mode[1] = MODE_ROS;
			
			if(c&0x20)
				mode[2] = MODE_RWS;
			else
				mode[2] = MODE_ROS;
				
			if(c&0x10)
				mode[3] = MODE_RWS;
			else
				mode[3] = MODE_ROS;
				
			break;
			
		case 1:	/* Common/Language ROS */
			if(!nxros_wtoggle)
			{
				nxros_addr = ((USHORT)c) << 8;
				nxros_wtoggle = 1;
			}
			else
			{
				nxros_addr |= (USHORT)c;
				nxros_wtoggle = 0;
			}
			nxros_rtoggle = 0;
			nxros_xtoggle = 0;
			break;

		case 5:	/* Printer */
			PrinterPUTB(c);
			break;
			
		case 13:	/* Diskette */
			DiskPUTB(c);
			break;
			
		case 14:	/* Tape */
			TapePUTB(c);
			break;
	}
}

UBYTE DoGETB(int dev)
{
	UBYTE val;
	
	switch(dev)
	{
		case 1:	/* Common/Language ROS */
#ifdef LITTLE_ENDIAN
			if(!nxros_rtoggle)
				val = nxros[nxros_addr] & 0xFF;
			else
			{
				val = nxros[nxros_addr] >> 8;
				nxros_addr++;
			}
#else
			if(!nxros_rtoggle)
				val = nxros[nxros_addr] >> 8;
			else
			{
				val = nxros[nxros_addr] & 0xFF;
				nxros_addr++;
			}
#endif
			nxros_rtoggle = 1-nxros_rtoggle;
			TRACE(val, 1);
			break;

		case 5:	/* Printer */
			val = PrinterGETB();
			break;
			
		case 13:	/* Diskette */
			val = DiskGETB();
			break;
			
		case 14:	/* Tape */
			val = TapeGETB();
			break;
			
		default:
			val = 0xFF;
	}

	return val;
}

void emu_keyboard(UBYTE code)
{
	int3 = TRUE;
	kbdcode = code;
}

void emu_dump_mem()
{
	FILE *outfile;
	
	outfile = fopen("mem.dmp", "wb");
	if(!outfile)
		return;
	fwrite(RWS, 1, 65536, outfile);
	fclose(outfile);
}

void emu_toggle_dump()
{
	if(!dump)
	{
		dumpfile = fopen("emu.dmp", "w");
		if(dumpfile)
			dump = 1;
	}
	else
	{
		dump = 0;
		fclose(dumpfile);
	}
}

void emu_select_lang(int lang)
{
	if(lang)
		language_ros = APLROS;
	else
		language_ros = BASROS;
}

int load_ros()
{
	int err;
	FILE *infile;
	size_t rossize;

	ExecROS = NULL;
	APLROS = NULL;
	BASROS = NULL;
	CommonROS = NULL;
	APLNXROS = NULL;
	BASNXROS = NULL;

	infile = fopen(EXECROS_NAME, "rb");
	if(!infile)
		return ERR_EXECROS_FILE;
	fseek(infile, 0, SEEK_END);
	rossize = ftell(infile);
	fseek(infile, 0, SEEK_SET);
	printf("Size of Executable ROS: %d\n", rossize);
	ExecROS = (USHORT *)malloc(rossize);
	if(!ExecROS)
		return ERR_EXECROS_ALLOC;
	if(fread(ExecROS, 1, rossize, infile) != rossize)
		return ERR_EXECROS_READ;
	fclose(infile);

	infile = fopen(COMMONROS_NAME, "rb");
	if(!infile)
		return ERR_COMMONROS_FILE;
	fseek(infile, 0, SEEK_END);
	rossize = ftell(infile);
	fseek(infile, 0, SEEK_SET);
	printf("Size of Common ROS: %d\n", rossize);
	CommonROS = (USHORT *)malloc(65536*2);
	if(!CommonROS)
		return ERR_COMMONROS_ALLOC;
	memset(CommonROS, 0, 65536*2);
	if(fread(CommonROS, 1, rossize, infile) != rossize)
		return ERR_COMMONROS_READ;
	fclose(infile);

	infile = fopen(BASNXROS_NAME, "rb");
	if(!infile)
		return ERR_BASNXROS_FILE;
	fseek(infile, 0, SEEK_END);
	rossize = ftell(infile);
	fseek(infile, 0, SEEK_SET);
	printf("Size of NX BASIC ROS: %d\n", rossize);
	BASNXROS = (USHORT *)malloc(65536*2);
	if(!BASNXROS)
		return ERR_BASNXROS_ALLOC;
	memset(BASNXROS, 0, 65536*2);
	if(fread(BASNXROS, 1, rossize, infile) != rossize)
		return ERR_BASNXROS_READ;
	fclose(infile);

	infile = fopen(BASROS_NAME, "rb");
	if(!infile)
		return ERR_BASROS_FILE;
	fseek(infile, 0, SEEK_END);
	rossize = ftell(infile);
	fseek(infile, 0, SEEK_SET);
	printf("Size of BASIC ROS: %d\n", rossize);
	BASROS = (USHORT *)malloc(rossize);
	if(!BASROS)
		return ERR_BASROS_ALLOC;
	if(fread(BASROS, 1, rossize, infile) != rossize)
		return ERR_BASROS_READ;
	fclose(infile);

	infile = fopen(APLNXROS_NAME, "rb");
	if(!infile)
		return ERR_APLNXROS_FILE;
	fseek(infile, 0, SEEK_END);
	rossize = ftell(infile);
	fseek(infile, 0, SEEK_SET);
	printf("Size of NX APL ROS: %d\n", rossize);
	APLNXROS = (USHORT *)malloc(65536*2);
	if(!APLNXROS)
		return ERR_APLNXROS_ALLOC;
	memset(APLNXROS, 0, 65536*2);
	if(fread(APLNXROS, 1, rossize, infile) != rossize)
		return ERR_APLNXROS_READ;
	fclose(infile);

	infile = fopen(APLROS_NAME, "rb");
	if(!infile)
		return ERR_APLROS_FILE;
	fseek(infile, 0, SEEK_END);
	rossize = ftell(infile);
	fseek(infile, 0, SEEK_SET);
	printf("Size of APL ROS: %d\n", rossize);
	APLROS = (USHORT *)malloc(rossize);
	if(!APLROS)
		return ERR_APLROS_ALLOC;
	if(fread(APLROS, 1, rossize, infile) != rossize)
		return ERR_APLROS_READ;
	fclose(infile);

	return 0;
}

int emu_init()
{
	int err;

	dump = 0;
	
	RWS = (USHORT *)malloc(65536);
	if(!RWS)
	{
		err = ERR_RWS_ALLOC;
		goto error;
	}
	RWSb = (UBYTE *)RWS;
		
	err = load_ros();

error:
	if(err)
		printf("ERROR: ");
	switch(err)
	{
		case ERR_RWS_ALLOC:
			printf("Could not allocate memory for Read/Write Storage\n");
			break;
		case ERR_EXECROS_FILE:
			printf("Could not open Executable ROS file %s\n", EXECROS_NAME);
			break;
		case ERR_EXECROS_ALLOC:
			printf("Could not allocate memory for Executable ROS\n");
			break;
		case ERR_EXECROS_READ:
			printf("Error reading Executable ROS\n");
			break;
		case ERR_COMMONROS_FILE:
			printf("Could not open Common ROS file %s\n", COMMONROS_NAME);
			break;
		case ERR_COMMONROS_ALLOC:
			printf("Could not allocate memory for Common ROS\n");
			break;
		case ERR_COMMONROS_READ:
			printf("Error reading Common ROS\n");
			break;
		case ERR_APLROS_FILE:
			printf("Could not open APL ROS file %s\n", APLROS_NAME);
			break;
		case ERR_APLROS_ALLOC:
			printf("Could not allocate memory for APL ROS\n");
			break;
		case ERR_APLROS_READ:
			printf("Error reading APL ROS\n");
			break;
		case ERR_APLNXROS_FILE:
			printf("Could not open APL NX ROS file %s\n", APLNXROS_NAME);
			break;
		case ERR_APLNXROS_ALLOC:
			printf("Could not allocate memory for APL NX ROS\n");
			break;
		case ERR_APLNXROS_READ:
			printf("Error reading APL NX ROS\n");
			break;
		case ERR_BASROS_FILE:
			printf("Could not open BASIC ROS file %s\n", BASROS_NAME);
			break;
		case ERR_BASROS_ALLOC:
			printf("Could not allocate memory for BASIC ROS\n");
			break;
		case ERR_BASROS_READ:
			printf("Error reading BASIC ROS\n");
			break;
		case ERR_BASNXROS_FILE:
			printf("Could not open BASIC NX ROS file %s\n", BASNXROS_NAME);
			break;
		case ERR_BASNXROS_ALLOC:
			printf("Could not allocate memory for BASIC NX ROS\n");
			break;
		case ERR_BASNXROS_READ:
			printf("Error reading BASIC NX ROS\n");
			break;
	}
	
	return err;
}

void emu_cleanup()
{
	halt = TRUE;

	if(RWS) free(RWS);
	if(APLROS) free(APLROS);
	if(APLNXROS) free(APLNXROS);
	if(BASROS) free(BASROS);
	if(BASNXROS) free(BASNXROS);
	if(CommonROS) free(CommonROS);
	if(ExecROS) free(ExecROS);
}
