/*
   emu5110
	-------
	An emulator for the IBM 5110
	(C) 2007-2008 Christian Corti
	
	Main program
*/

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <limits.h>
#include <time.h>
#include <stdlib.h>
#include <pthread.h>
#include <readline/readline.h>
#include <readline/history.h>

#include "emu.h"

pthread_mutex_t emu_mutex;
pthread_cond_t emu_cond_run;

int quit;
int run;

void LoadFile(char *cmdline)
{
	FILE *infile;
	char *fname, *base;
	unsigned long baseval;

	fname = strchr(cmdline, ' ');
	if(!fname)
	{
		printf("No file specified.\n");
		return;
	}
	fname++;

	base = strchr(fname, ' ');
	if(!base)
		baseval = 0x0B00;
	else
	{
		*base++ = '\0';
		baseval = strtol(base, NULL, 0);
	}
		
	if(baseval < 0x200)
	{
		printf("Illegal load address %lu.\n", baseval);
		return;
	}
	
	infile = fopen(fname, "rb");
	if(!infile)
	{
		fprintf(stderr, "Can't open '%s': ", fname);
		perror(NULL);
		return;
	}
	fread(&RWSb[baseval], 65536-baseval, 1, infile);
	fclose(infile);
	printf("Done.\n");
}

void Halt()
{
	pthread_mutex_lock(&emu_mutex);
	run = 0;
	pthread_cond_signal(&emu_cond_run);
	pthread_mutex_unlock(&emu_mutex);
}

void Run()
{
	pthread_mutex_lock(&emu_mutex);
	run = 1;
	emu_start();
	pthread_cond_signal(&emu_cond_run);
	pthread_mutex_unlock(&emu_mutex);
}

void *DisplayThread(void *threadid)
{
	struct timespec ts;

	while(!quit)
	{
		if(DoEvents())
			continue;
#if 0
		usleep(50000);		usleep is not thread-safe on Solaris
#else
		ts.tv_sec = 0;
		ts.tv_nsec = 50000000L;	/* 50ms */
		nanosleep(&ts, NULL);
#endif
		UpdateScreen();
	}
	pthread_exit(0);
}

void *EmuThread(void *threadid)
{
	while(!quit)
	{
		pthread_mutex_lock(&emu_mutex);
		if(!run)
			pthread_cond_wait(&emu_cond_run, &emu_mutex);
		pthread_mutex_unlock(&emu_mutex);
		
		if(run)
		{
			if(emu_do() == TRUE)
				Halt();
		}
	}
	pthread_exit(0);
}

int main(int argc, char *argv[])
{
	pthread_t threads[10];
	int num_threads=0;
	int i, retval;
	char *cmd;

	/* Initialize emulator and I/O */

	if(emu_init())
		return 10;
	emu_select_lang(0);
	emu_reset();

	if(DisplayInit())
		return 20;

	if(!PrinterAttach())
	{
		fprintf(stderr, "ERROR: Can't open printer output file\n");
		return 21;
	}		
	DiskSetup();
	TapeSetup();

	quit = 0;
	run = 0;

	pthread_mutex_init(&emu_mutex, NULL);
	pthread_cond_init(&emu_cond_run, NULL);

	/* Create and start threads */

	retval = pthread_create(&threads[0], NULL, EmuThread, (void *)num_threads++);
	if(retval)
	{
		fprintf(stderr, "ERROR: Can't create emulation thread\n");
		return 30;
	}
	retval = pthread_create(&threads[1], NULL, DisplayThread, (void *)num_threads++);
	if(retval)
	{
		fprintf(stderr, "ERROR: Can't create display thread\n");
		return 31;
	}

	/* MAIN LOOP */

	using_history();
	while(!quit)
	{
		cmd = readline("emu5110> ");
		add_history(cmd);
		switch(cmd[0])
		{
			case 'a':
				if(!strncasecmp(cmd, "apl", 3))
				{
					Halt();
					emu_select_lang(1);
					emu_reset();
					Run();
				}
				else if(!strncasecmp(cmd, "attach", 3))
				{
					char *cc, name[PATH_MAX+1];
					int drv;

					cc = strchr(cmd, ' ');
					if(!cc)
					{
						printf("Syntax: ATTACH drive filename\n");
						break;
					}
					cc++;
					sscanf(cc, "%d %s", &drv, name);
					if(drv<1 || drv>6)
					{
						printf("Invalid drive number %d, use 1..4 for D80..D10 and 5..6 for E80..E40\n", drv);
						break;
					}
					if(!name[0])
					{
						printf("No filename specified.\n");
						break;
					}
					
					if(drv < 5)
					{
						if(!DiskAttach(--drv, name))
							perror("ATTACH failed for disk");
					}
					else
					{
						if(!TapeAttach(drv-5, name))
							perror("ATTACH failed for tape");
					}
				}
				break;

			case 'b':
				if(!strncasecmp(cmd, "basic", 3))
				{
					Halt();
					emu_select_lang(0);
					emu_reset();
					Run();
				}
				break;
				
			case 'd':
				if(!strncasecmp(cmd, "detach", 3))
				{
					char *cc, name[PATH_MAX+1];
					int drv;

					cc = strchr(cmd, ' ');
					if(!cc)
					{
						printf("Syntax: DETACH drive\n");
						break;
					}
					cc++;
					sscanf(cc, "%d", &drv);
					if(drv<1 || drv>6)
					{
						printf("Invalid drive number %d, use 1..4 for D80..D10 and 5..6 for E80..E40\n", drv);
						break;
					}
					
					if(drv < 5)
						DiskDetach(--drv);
					else
						TapeDetach(drv-5);
				}
				else if(!strncasecmp(cmd, "dump", 2))
				{
					emu_toggle_dump();
				}
				break;
				
			case 'g':
				Run();
				break;

			case 'h':
				Halt();
				break;
				
			case 'l':
				LoadFile(cmd);
				break;
				
			case 'm':
				emu_dump_mem();
				break;

			case 'q':
				Halt();
				quit = 1;
				break;
				
			case 'r':
				emu_reset();
				break;
				
			case 's':
				Halt();
				emu_fetch();
				dump_regs();
				break;
				
		}
		free(cmd);
	}	

	pthread_cond_broadcast(&emu_cond_run);
	for(i=0; i<num_threads; i++)
		pthread_join(threads[i], NULL);

	PrinterDetach();

	return 0;
}
